import numpy as np
from .util.math import range_step
from .util.functions import compose

_distribution_samples = {
    'float': {
        'normal': compose(False,
                lambda **kw: np.random.normal(kw['mean'], kw['std'], kw['size'])
        ),
        'uniform': compose(False,
            lambda **kw: np.random.uniform(kw['min'], kw['max'], kw['size'])
        ),
        'range': compose(False,
            lambda **kw: np.arange(kw['start'], kw['end'], range_step(kw['start'], kw['end'], kw['size'])),
            lambda f: lambda **kw: f(**kw).astype(float)[:kw['size']]
        )
    },
    'integer': {
        'uniform': compose(False,
            lambda **kw: np.random.random_integers(kw['min'], kw['max'], kw['size'])
        ),
        'binomial': compose(False,
            lambda **kw: np.random.binomial(kw['n'], kw['p'], kw['size'])
        ),
        'range': compose(False,
            lambda **kw: np.arange(kw['start'], kw['end'], range_step(kw['start'], kw['end'], kw['size'])),
            lambda f: lambda **kw: f(**kw).astype(int)[:kw['size']]
        )
    }
}


TYPES = { 'float', 'integer' }


DISTRIBUTIONS = {
    'normal': { 'mean', 'std' },
    'uniform': { 'min', 'max' },
    'binomial' : { 'n', 'p' },
    'range': { 'start', 'end' }
}


def get_sample(dtype: str, size: int = None, **props):
    size = size or props['size']
    distr = props['distr']
    return _distribution_samples[dtype][distr]( **(props | {'size': size}) )



mocker_config = {}


def config(**kwargs):
    mocker_config.update(kwargs)


def get_configs(*keys):
    if len(list(keys)) == 0:
        keys = mocker_config.keys()
    return (config.get(key) for key in keys)



import os.path as osp
from pandas import read_pickle, read_csv, DataFrame
from pathlib import Path
import json


class _Datacache:

    data = {}

    def __init__(self):
        return

    def write(self, key: str, data, **kwargs):
        self.data[key] = { 'data': data }
        self.data[key].update(kwargs)

    def read(self, key: str):
        if not self.data.get(key):
            return None
        return self.data[key]['data']

    def get_props(self, key: str, *props):
        return { key: self.data[key][kw] for kw in props }


_cache = _Datacache()


_df_writers = {
    '.pkl' : lambda file, dataframe: dataframe.to_pickle(file, index=True),
    '.csv' : lambda file, dataframe: dataframe.to_csv(file, index=False),
    '.tsv' : lambda file, dataframe: dataframe.to_csv(file, sep="\t", index=False),
    '.json' : lambda file, dataframe: dataframe.to_json(file, orient='table', index=False)
}


get_dataset_path = lambda dataset: osp.join(osp.dirname(__file__), 'data', dataset + '.pkl')


DATASETS = { 'first_name', 'last_name' }


def load_dataset(dataset: str):
    path = get_dataset_path(dataset)
    data = _cache.read(path)
    if data is None:
        data = read_pickle(path)
        _cache.write(path, data)
    return data


def load_table(path: str):
    data = _cache.read(path)
    if data is None:
        data = read_csv(path)
        _cache.write(path, data)
    return data


def write_dataframe(file: str, dataframe: DataFrame):
    file_ext = Path(file).suffix
    if file_ext is None or not file_ext in _df_writers.keys():
        file_ext = '.csv'
    _df_writers[file_ext](file, dataframe)


def load_json(file) -> dict:
    f = open(file, 'rt') if type(file) is str else file
    return json.load(f)


def write_json(obj: dict, path: str, pretty: str, indent: int):
    with open(path, 'wt', ) as f:
        json.dump(obj, f, indent=(indent if pretty else None))



from .io import load_dataset, load_table, DATASETS
from pandas import Series
from .numbers import get_sample as num_sample, TYPES as NUMTYPES
from .time import get_sample as time_sample


def _table_column_sample(path: str, field_name: str):
    dataset = load_table(path)
    return lambda size: [dataset[field_name].sample(n=size, replace=True).reset_index(drop=True)]


def get_sample_generators(field_name: str, mock_type: str, **props):
    if mock_type in DATASETS:
        return lambda size: load_dataset(mock_type).sample(n=size, ignore_index=True, replace=True)
    elif mock_type in NUMTYPES:
        return lambda size: num_sample(mock_type, size, **props)
    elif mock_type == 'enum':
        return lambda size: Series(props['values']).sample(n=size, ignore_index=True, replace=True, weights=props['weights'])
    elif mock_type in { 'date', 'datetime' }:
        return lambda size: time_sample(mock_type, size, **props)
    elif mock_type == 'table':
        return _table_column_sample(props['path'], field_name)
    else:
        raise ValueError('Unsupported type')



import pandas as pd
from numpy import arange
from .util.functions import compose


ISO_DATETIME = '%Y-%m-%dT%H:%M%:%SZ'

ISO_DATE = '%Y-%m-%d'


default_format = {
    'date': ISO_DATE,
    'datetime': ISO_DATETIME
}


_distribution_samples = {
    'range': compose(False,
        lambda **kw: pd.date_range(start=kw['start'], end=kw['end'], periods=kw['size']),
        lambda f: lambda **kw: f(**kw).strftime(default_format[kw['type']]),
        lambda f: lambda **kw: f(**kw).to_series(index=arange(kw['size']))
    ),
}
_distribution_samples['uniform'] = (
    compose(False,
        lambda **kw: pd.date_range(_distribution_samples['range'])(**kw),
        lambda f: lambda **kw: f(**kw).sample(frac=1).reset_index(drop=True)
    )
)


def get_sample(mock_type: str, size: int, **kw):
    kw['type'] = mock_type
    distr = kw['distr']
    return _distribution_samples[distr](**( kw | {'size' : size} ) )



from pandas import Series, DataFrame, concat
from .sampler import get_sample_generators
from .config import get_configs

# def build_series(name: str, field_descriptor: dict, size: int):
    # get_sample = get_sample_generators(field_descriptor['mock_type'], **field_descriptor['props'])
    # return Series(data=get_sample(size), name=name)


def build_dataframe(fields_describe: dict, size: int) -> DataFrame:
    df = None
    report_progress = get_configs('report_progress')
    for field, field_spec in fields_describe['fields'].items():
        report_progress and print('Sampling', field, '...')
        sample = get_sample_generators(field, field_spec['mock_type'], **field_spec['props'])(size)
        if not type(sample) is Series:
            name = field
        else:
            name = sample.name or field
        df = concat([df, Series(sample, name=name)], axis=1)
    report_progress and print('Done!')
    return df



class UnsupportedMockTypeException(Exception):

    def __init__(self, mock_type: str):
        msg = f"Kind '{mock_type}' is not supported"
        super().__init__(msg)


class SchemaPropValueException(Exception):

    def __init__(self, prop, value, rule):
        fmtd_rule = str(rule).lower().strip(' ')
        msg = f"Invalid property value for '{prop}'. Value: '{value}' {fmtd_rule}"
        super().__init__(msg)


class SchemaRepeatedFieldException(Exception):

    def __init__(self, field_name):
        msg = f"Repeated field: '{field_name}'"
        super().__init__(msg)


class MissingPropsException(Exception):

    def __init__(self, mock_type: str, missing_props: set):
        msg = f"Missing required fields for kind '{mock_type}': {str(missing_props)}"
        super().__init__(msg)



from pandas import DataFrame
from .io import write_dataframe, load_json, write_json, load_table
from .builder import build_dataframe
from .util.list import dedup_list, list_diff


def _config_column_order(specified, fields_dict):
    if not specified or len(specified) < 1:
        return fields_dict
    spec_dedup = dedup_list(specified)
    order = spec_dedup + list_diff(fields_dict.keys(), spec_dedup)
    return { key: fields_dict[key] for key in order }


class MockGenerator:

    dataframe: DataFrame
    fields_describe = {
        'fields': dict()
    }

    def __init__(self, config = None) -> None:
        if config:
            self.fields_describe = load_json(config)

    def dump_config(self, path, pretty=True, indent=2):
        write_json(self.fields_describe, path, pretty, indent)

    def add_field(self, name: str, mock_type:str, **props):
        self.fields_describe['fields'][name] = {
            'mock_type': mock_type,
            'props': props
        }

    def add_fields(self, fields_dict: dict):
        self.fields_describe['fields'].update(fields_dict)

    def add_table(self, path: str):
        columns = load_table(path).columns
        for col in columns:
            self.fields_describe['fields'][col] = {'mock_type': 'table', 'props': { 'path': path }}

    def sample(self, size: int):
        self.dataframe = build_dataframe(self.fields_describe, size)

    def get_dataframe(self):
        return self.dataframe.copy()

    def dump(self, path):
        if not path:
            raise ValueError('Path must be specified')
        write_dataframe(path, self.dataframe)

    def set_column_order(self, order):
        self.fields_describe['fields'] = _config_column_order(order, self.fields_describe['fields'])

    def __str__(self) -> str:
        self.dataframe.__str__()

    def __repr__(self) -> str:
        self.dataframe.__str__()
